 /*
  * XMMS Visualization Plugin by Pascal Brochart
  * --------------- Plazma 0.0.7 ---------------
  * -------   Copyright (C) 10/25/2001   -------
  * ---- E-mail: p.brochart@libertysurf.fr -----
 */



#include <gtk/gtk.h>
#include <math.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#include <xmms/plugin.h>
#include <xmms/util.h>
#include <stdio.h>
#include <stdlib.h>

#ifndef PI
#define PI	3.14159265358979323846
#endif
#define WIDTH 489
#define HEIGHT 400
#define TABLEX WIDTH*2
#define TABLEY HEIGHT*2
#define TONES 512
#define SDL_FLAGS SDL_HWSURFACE|SDL_HWPALETTE|SDL_DOUBLEBUF|SDL_ASYNCBLIT|SDL_HWACCEL|SDL_RLEACCEL
#define RESFACTXF(par1) ( (float) resx*(par1)/WIDTH )
#define RESFACTYF(par1) ( (float) resy*(par1)/HEIGHT )

short pcm_buffer[1024];
unsigned short render_buffer[512];
unsigned short val_maxi=127;
gboolean changement=FALSE, in_thread=FALSE, finished=FALSE;
int i, c,chcol0=36, pitch, bass, mouse, fullscreen, quit_timeout;
int state=1368, old_state=1368, resx=WIDTH, resy=HEIGHT;
guint8 *pixel;
double r[2], R[4];
float compt_grille_3d, compt_cercle_3d;
SDL_Color colors[TONES*2];
unsigned char *tmem, *tmem1;
SDL_Surface *ecran_sdl;
SDL_Thread *plazma_thread;
SDL_mutex *acces_ressource;
SDL_Event plazma_event;
static GtkWidget *about_fenetre, *boite_verticale, *fenetre_config;
static GtkWidget *boite_config1, *boite_config2, *about_inscription;
static GtkWidget *about_inscription, *about_bouton, *close, *bouton_config;
static GtkWidget *bouton_config2, *ok, *cancel, *frame, *hbox, *scroller;
static GtkWidget *hbox_scroll, *frame_scroll, *vbox_scroll, *label_scroll;
static GtkWidget *bouton_lines, *bouton_spectrum, *bouton_3d, *frame_options;
static GtkWidget *vbox_options, *rot_lines, *radial_lines;
static GSList *group, *group_options;
static GtkObject *dimension;

static void plazma_init(void);
static void plazma_about(void);
static void plazma_config(void);
static void plazma_about_quit(GtkWidget *w, GtkWidget **about_fenetre);
static void config_ok(GtkWidget *w, GtkWidget *window);
static void config_cancel(GtkWidget *w, GtkWidget *window);
static void plazma_standard_effect(void);
static void plazma_artistic_effect(void);
static void plazma_scroll_changed(GtkWidget *w, int *changed_value);
static void plazma_cleanup(void);
static void plazma_render_freq(short spectrum[2][256]);
static void plazma_render_pcm(short pcm[2][512]);
static void func_thread(void);
static void plazma_change_effect(void);
static void plazma_check_effect_la(void);
static void plazma_options_rot(void);
static void plazma_options_radial(void);
static void plazma_end(void);
static void plazma_create(void);

struct sincos {
	int i;
	float *f;
};

struct general {
	int effect;
	int lines;
	int spectrum;
	int use_3d;
	int options;
	int bass_sensibility;
	int last_effect;
	int last_lines;
	int last_use_3d;
	int last_bass_sensibility;
	float k_put;
	float rot_tourni;
};

static struct sincos cosw = { 0, NULL };
static struct sincos sinw = { 0, NULL };
static struct general my_effect = { TRUE, TRUE, TRUE, TRUE, 0, 0, 0, 0, 0, 0, 0, 0.4 };
static struct general *ptr_effect = &my_effect;

VisPlugin plazma_plugin =
{
	NULL,
	NULL,
	0,
	"Plazma 0.0.7",
	1,
	1,		
	plazma_init, /* init */
	plazma_cleanup, /* cleanup */
	plazma_about, /* about */
	plazma_config, /* configure */
	NULL, /* disable_plugin */
	NULL, /* playback_start */
	NULL, /* playback_stop */
	plazma_render_pcm, /* render_pcm */
	plazma_render_freq  /* render_freq */
};

VisPlugin *get_vplugin_info(void)
{
   return &plazma_plugin;
}

static void do_plasma(SDL_Surface*surface, double x1, double y1,
                                           double x2, double y2,
                                           unsigned char*t)
{
	unsigned int	X1=x1*(TABLEX/2), Y1=y1*(TABLEY/2),
			X2=x2*(TABLEX/2), Y2=y2*(TABLEY/2), y;
	unsigned char 	*t1=t+X1+Y1*TABLEX, *t2=t+X2+Y2*TABLEX;
	for (y=0; y<HEIGHT; y++) {
		unsigned char*tmp = pixel + y * pitch;
		unsigned int t = y*TABLEX, tmax=t+WIDTH; 
		for (t=t; t<tmax; t++, tmp++)
			tmp[0]=t1[t]+t2[t];
	}
}

int timeout_fn(gpointer data)
{
	if (in_thread)	{
		plazma_plugin.disable_plugin(&plazma_plugin);
		printf("Disable plugin...\n");
	}
}

static void plazma_init(void)
{
	fullscreen = FALSE;
	finished = FALSE;
	in_thread = FALSE;
	for (c=0; c<2; c++)
		r[c]=((double)(cos(c)*120))/5000;
	for (c=0; c<4; c++)
		R[c]=((double)(cos(c)*120))/5000;
	printf("\nPlazma initialisation:\n");
	tmem = (unsigned char *)malloc(TABLEY*TABLEX);
	tmem1 = (unsigned char *)malloc(TABLEY*TABLEX);
	if (tmem == NULL || tmem1 == NULL)
		printf("Not enought memory !!!\n");
	else
		printf("Creating buffer...\n");
	if (SDL_Init(SDL_INIT_VIDEO) < 0)	{
		printf("Can't initialize SDL !!!\n");
		exit(1);
	}
	printf("Initialize SDL...\n");
	ecran_sdl = SDL_SetVideoMode(WIDTH, HEIGHT, 8, SDL_FLAGS);
	if (ecran_sdl == NULL)	{
		printf("Can't open SDL screen !!!\n");
		exit(1);
	}
	else
		printf("Open SDL screen...\n");
	pitch = ecran_sdl->pitch;
	pixel = (Uint8*)ecran_sdl->pixels;
	SDL_WM_SetCaption("Plazma XMMS Plugin",NULL);	
	plazma_create();
	plazma_change_effect();	
	acces_ressource = SDL_CreateMutex();
	SDL_SetColors(ecran_sdl, colors, 0, TONES);	
	plazma_thread = SDL_CreateThread ((void *) func_thread, NULL);
	quit_timeout = gtk_timeout_add (1000, timeout_fn, NULL);
}

static void plazma_about(void)
{
	if (about_fenetre)
		return;
	about_fenetre = gtk_window_new(GTK_WINDOW_DIALOG);
	boite_verticale = gtk_vbox_new(FALSE, 4);
	gtk_window_set_title(GTK_WINDOW(about_fenetre), "About Plazma");
	gtk_window_set_policy(GTK_WINDOW(about_fenetre), FALSE, FALSE, FALSE);
	gtk_window_set_position(GTK_WINDOW(about_fenetre), GTK_WIN_POS_MOUSE);
   	gtk_signal_connect(GTK_OBJECT(about_fenetre), "destroy", 
GTK_SIGNAL_FUNC(gtk_widget_destroyed), &about_fenetre);
	gtk_widget_set_usize(about_fenetre, 320, 210);
	gtk_container_add(GTK_CONTAINER(about_fenetre), boite_verticale);
	gtk_container_set_border_width(GTK_CONTAINER(boite_verticale),8);
	gtk_widget_realize(about_fenetre);
	gtk_widget_show(boite_verticale);
	about_inscription = gtk_label_new ("Plazma 0.0.7\n\n\
Copyright (C) 2000-2001 Pascal Brochart\n Homepage: \
<http://diablolight.multimania.com>\n\
E-mail: <p.brochart@libertysurf.fr>\n\n\
This program is free software, you can redistribute it\n\
under the terms of the GNU General Public License.");
	gtk_box_pack_start(GTK_BOX(boite_verticale), about_inscription, FALSE, FALSE, 
8);
	gtk_widget_show(about_inscription);
	about_bouton = gtk_hbutton_box_new();
	gtk_box_pack_end(GTK_BOX(boite_verticale), about_bouton, FALSE, FALSE, 8);
	close = gtk_button_new_with_label("Close");
	GTK_WIDGET_SET_FLAGS(close, GTK_CAN_DEFAULT);
	gtk_window_set_default(GTK_WINDOW(about_fenetre), close);	/* enable */
	gtk_hbutton_box_set_layout_default(GTK_BUTTONBOX_END);	/* bas-droite */
	gtk_box_pack_end(GTK_BOX(about_bouton), close, FALSE, FALSE, 8);
	gtk_widget_show (about_bouton);
	gtk_widget_show (close);
	gtk_signal_connect(GTK_OBJECT(close), "clicked",
		GTK_SIGNAL_FUNC(plazma_about_quit), &about_fenetre);
	gtk_widget_show(about_fenetre);
}

static void plazma_about_quit(GtkWidget *w, GtkWidget **about_fenetre)
{
	gtk_widget_destroy(*about_fenetre);
	*about_fenetre = NULL;
}

static void plazma_config(void)
{
	if (fenetre_config)
		return;
	ptr_effect->last_effect = ptr_effect->effect;
	ptr_effect->last_bass_sensibility = ptr_effect->bass_sensibility;
	ptr_effect->last_lines = ptr_effect->lines;
	ptr_effect->last_use_3d = ptr_effect->use_3d;
	fenetre_config = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_title (GTK_WINDOW(fenetre_config), "Plazma Configuration");
	gtk_window_set_policy(GTK_WINDOW(fenetre_config), FALSE, FALSE, FALSE);
	gtk_window_set_position(GTK_WINDOW(fenetre_config), GTK_WIN_POS_MOUSE);
	gtk_widget_set_usize(fenetre_config, 370, 352);
	gtk_container_set_border_width (GTK_CONTAINER(fenetre_config),10);
   	gtk_signal_connect(GTK_OBJECT(fenetre_config), "destroy", 
GTK_SIGNAL_FUNC(gtk_widget_destroyed), &fenetre_config);
	frame = gtk_frame_new("Background");
	boite_config1 = gtk_vbox_new (FALSE,0);
	gtk_container_add (GTK_CONTAINER (fenetre_config), boite_config1);
	gtk_box_pack_start (GTK_BOX(boite_config1), frame, TRUE, TRUE, 4);	
	boite_config2 = gtk_vbox_new (FALSE,0);
	gtk_container_set_border_width (GTK_CONTAINER (boite_config2), 5);
	gtk_container_add (GTK_CONTAINER (frame), boite_config2);
	bouton_config = gtk_radio_button_new_with_label (NULL, "Standard effect");
	if (ptr_effect->effect == FALSE)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (bouton_config), TRUE);
	gtk_box_pack_start (GTK_BOX (boite_config2), bouton_config, FALSE, FALSE, 0);
	group = gtk_radio_button_group (GTK_RADIO_BUTTON (bouton_config));
	bouton_config2 = gtk_radio_button_new_with_label (group, "Artistic effect");
	if (ptr_effect->effect == TRUE)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (bouton_config2), TRUE);
	gtk_box_pack_start (GTK_BOX (boite_config2), bouton_config2, FALSE, FALSE, 0);
	gtk_signal_connect (GTK_OBJECT (bouton_config), "toggled",
		GTK_SIGNAL_FUNC(plazma_standard_effect), fenetre_config);
	gtk_signal_connect (GTK_OBJECT (bouton_config2), "toggled",
		GTK_SIGNAL_FUNC(plazma_artistic_effect), fenetre_config);

	frame_options = gtk_frame_new("Effects");
	gtk_box_pack_start (GTK_BOX(boite_config1), frame_options, TRUE, TRUE, 4);
	vbox_options = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (vbox_options), 5);
	gtk_container_add (GTK_CONTAINER (frame_options), vbox_options);
	rot_lines = gtk_radio_button_new_with_label (NULL, "Rotation of grids");
	if (ptr_effect->options == FALSE)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (rot_lines), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox_options), rot_lines, FALSE, FALSE, 0);
	group_options = gtk_radio_button_group (GTK_RADIO_BUTTON (rot_lines));
	radial_lines = gtk_radio_button_new_with_label (group_options, "Wave radial");
	if (ptr_effect->options == TRUE)
		gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (radial_lines), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox_options), radial_lines, FALSE, FALSE, 0);
	gtk_signal_connect (GTK_OBJECT (rot_lines), "toggled",
		GTK_SIGNAL_FUNC(plazma_options_rot), fenetre_config);
	gtk_signal_connect (GTK_OBJECT (radial_lines), "toggled",
		GTK_SIGNAL_FUNC(plazma_options_radial), fenetre_config);

	frame_scroll = gtk_frame_new("General");
	gtk_box_pack_start (GTK_BOX(boite_config1), frame_scroll, TRUE, TRUE, 4);
	vbox_scroll = gtk_vbox_new (FALSE, 0);
	gtk_container_set_border_width (GTK_CONTAINER (vbox_scroll), 5);
	gtk_container_add (GTK_CONTAINER (frame_scroll), vbox_scroll);
	hbox_scroll = gtk_hbox_new(FALSE, 10);
	gtk_box_pack_start (GTK_BOX (vbox_scroll), hbox_scroll, TRUE, TRUE, 4);
	label_scroll = gtk_label_new ("Bass sensibility");
	gtk_box_pack_start (GTK_BOX(hbox_scroll), label_scroll, FALSE, FALSE, 3);
	dimension = gtk_adjustment_new (ptr_effect->bass_sensibility, -10, 10, 0, 0, 0);
	scroller = gtk_hscale_new (GTK_ADJUSTMENT(dimension));
	gtk_scale_set_draw_value (GTK_SCALE(scroller), TRUE);
	gtk_scale_set_value_pos (GTK_SCALE(scroller), GTK_POS_LEFT);
	gtk_scale_set_digits (GTK_SCALE(scroller), 0);
	gtk_box_pack_start (GTK_BOX (hbox_scroll), scroller, TRUE, TRUE, 2);
	bouton_3d = gtk_check_button_new_with_label("Active 3D rendering");
	if (ptr_effect->use_3d)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bouton_3d), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox_scroll), bouton_3d, FALSE, FALSE, 0);
	bouton_lines = gtk_check_button_new_with_label("Added oscilloscope");
	if (ptr_effect->lines)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bouton_lines), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox_scroll), bouton_lines, FALSE, FALSE, 0);
	bouton_spectrum = gtk_check_button_new_with_label("Added spectrum analyzer");
	if (ptr_effect->spectrum)
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bouton_spectrum), TRUE);
	gtk_box_pack_start (GTK_BOX (vbox_scroll), bouton_spectrum, FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(dimension), "value-changed", 
		GTK_SIGNAL_FUNC(plazma_scroll_changed), &ptr_effect->bass_sensibility);
	gtk_signal_connect (GTK_OBJECT (bouton_lines), "toggled",
		GTK_SIGNAL_FUNC(plazma_check_effect_la),fenetre_config);
	gtk_signal_connect (GTK_OBJECT (bouton_spectrum), "toggled",
		GTK_SIGNAL_FUNC(plazma_check_effect_la), fenetre_config);
	gtk_signal_connect (GTK_OBJECT (bouton_3d), "toggled",
		GTK_SIGNAL_FUNC(plazma_check_effect_la), fenetre_config);
	hbox = gtk_hbutton_box_new();
	gtk_button_box_set_layout (GTK_BUTTON_BOX(hbox), GTK_BUTTONBOX_END);
	gtk_button_box_set_spacing (GTK_BUTTON_BOX(hbox), 5);
	gtk_box_pack_start (GTK_BOX (boite_config1), hbox, FALSE, FALSE, 0);
	ok = gtk_button_new_with_label ("Ok");
	gtk_signal_connect (GTK_OBJECT (ok), "clicked",
		GTK_SIGNAL_FUNC(config_ok), fenetre_config);
	gtk_box_pack_start (GTK_BOX (hbox), ok, TRUE, TRUE, 0);
	GTK_WIDGET_SET_FLAGS (ok, GTK_CAN_DEFAULT);
	gtk_widget_grab_default (ok);	/* ok par defaut */
	cancel = gtk_button_new_with_label ("Cancel");
	gtk_signal_connect (GTK_OBJECT (cancel), "clicked",
		GTK_SIGNAL_FUNC(config_cancel), fenetre_config);
	gtk_box_pack_start (GTK_BOX (hbox), cancel, TRUE, TRUE, 0);
        GTK_WIDGET_SET_FLAGS (cancel, GTK_CAN_DEFAULT);
	gtk_widget_show_all (fenetre_config);
}

static void config_ok(GtkWidget *w, GtkWidget *window)
{
	gtk_widget_destroy (window);
	fenetre_config = NULL;
}

static void config_cancel(GtkWidget *w, GtkWidget *window)
{
	ptr_effect->effect = ptr_effect->last_effect;
	ptr_effect->bass_sensibility = ptr_effect->last_bass_sensibility;
	ptr_effect->lines = ptr_effect->last_lines;
	changement = TRUE;
	gtk_widget_destroy (window);
	fenetre_config = NULL;
}

static void plazma_standard_effect(void)
{
	ptr_effect->effect = FALSE;
	if (ecran_sdl == NULL)
		return;
	changement = TRUE;
}

static void plazma_artistic_effect(void)
{
	ptr_effect->effect = TRUE;
	if (ecran_sdl == NULL)
		return;
	changement = TRUE;
}

static void plazma_options_rot(void)
{
	ptr_effect->options = FALSE;
}

static void plazma_options_radial(void)
{
	ptr_effect->options = TRUE;
}

static void plazma_check_effect_la(void)
{
	ptr_effect->lines = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bouton_lines));
	ptr_effect->spectrum = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bouton_spectrum));
	ptr_effect->use_3d = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bouton_3d));
}

static void plazma_scroll_changed(GtkWidget *w, int *changed_value)
{
	*changed_value = (int) GTK_ADJUSTMENT(w)->value;
}

static void plazma_cleanup(void)
{
	if (in_thread)	{
		SDL_WaitThread(plazma_thread, NULL);
		gtk_timeout_remove (quit_timeout);
		return;
	}
	gtk_timeout_remove (quit_timeout);
	finished = TRUE;
	SDL_WaitThread(plazma_thread, NULL);
	plazma_thread = NULL;
	SDL_FreeSurface(ecran_sdl);
	free(tmem1);
	free(tmem);
	printf("Clear memory...\n");
	SDL_Quit();
	printf("Close SDL screen...\n");
}

static void plazma_change_effect(void)
{
	switch (ptr_effect->effect)	{
		case 0:
     		for (c=0 ; c<TONES; c++) {
		colors[c].r=(sin(((double)c)/TONES*6*PI+(sin(445)))+1)*127;
		colors[c].g=(sin(((double)c)/TONES*6*PI+(sin(561)))+1)*127;
		colors[c].b=(cos(((double)c)/TONES*6*PI+(sin(278)))+1)*127;
			}
		break;
		case 1:
    		for (c=0 ; c<TONES; c++) {
		colors[c].r=(sin(((double)c)/TONES*6*PI+(sin(c/4)))+1)*127;
		colors[c].g=(sin(((double)c)/TONES*6*PI+(sin(561)))+1)*127;
		colors[c].b=(cos(((double)c)/TONES*6*PI+(sin(278)))+1)*127;
			}
		break;
	}
}

static void plazma_create(void)
{		
    	int y;
  	for (y=0 ; y<TABLEY ; y++) {
   	  	int x;      		
		for (x=0 ; x<TABLEX ; x++) {
			double tmp = (((double)((x-(TABLEX/2))*(x-(TABLEX/2))+(y-(TABLEX/2))*(y-(TABLEX/2))))*(PI/(TABLEX*TABLEX+TABLEY*TABLEY)));
			tmem[y*TABLEX+x] = tan(tan(tmp*1500))+(tan(tmp)*40)+sqrt(exp(tmp)*10)+sin(tmp*10);
			tmem1[y*TABLEX+x] = sin(exp(tmp*35)*120)+sqrt(tmp*14)+log(exp(tmp*8)+cos(tmp*tmp))+tan(exp(tmp*40)*150)/6;
		}
   	}
}

static void plazma_render_pcm(short pcm[2][512])
{
	SDL_mutexP(acces_ressource);
	memcpy(pcm_buffer,pcm,1024);
	SDL_mutexV(acces_ressource);
}

static void plazma_render_freq(short spectrum[2][256])
{
	bass = 0;
	for (c=0 ; c<6 ; c++)
		bass += spectrum[0][c]>>4;
	old_state = state;
	if (ptr_effect->bass_sensibility >= 0)
		state += (bass/400)+1+(ptr_effect->bass_sensibility/2);
	if (ptr_effect->bass_sensibility < 0)	{
		state += (bass/400)+1-(abs(ptr_effect->bass_sensibility/2));
		if (state < (old_state+1))
			state = old_state+1;
	}
	SDL_mutexP(acces_ressource);
	memcpy(render_buffer,spectrum,512);
	SDL_mutexV(acces_ressource);
}

void aff_pixel(int x, int y, int colpix)
{
	unsigned char*point, *old_point;
	if (x<0 || x>(WIDTH-1) || y<0 || y>(HEIGHT-1))
		return;
	point = pixel+y*pitch;
	old_point = point;
	point[x] = old_point[x] | colpix;
	if (!ptr_effect->use_3d)
		point[x] = colpix;
}

void cercle (gint h, gint k, gint y, guint8 colcer)
{
	gint x = -1, d = 3 - 2 * y;
	while (x <= y)
	{
		if (d < 0)
		{
			d += 4 * x + 6;
			x++;
		}
		else
		{
			d += 4 * (x - y) + 10;
			x++;
			y--;
		}
		aff_pixel (h + x, y + k, colcer);
		aff_pixel (h + y, x + k, colcer);
		aff_pixel (h - y, x + k, colcer);
		aff_pixel (h - x, y + k, colcer);
		aff_pixel (h - x, -y + k, colcer);
		aff_pixel (h - y, -x + k, colcer);
		aff_pixel (h + y, -x + k, colcer);
		aff_pixel (h + x, -y + k, colcer);
	}
}

void droite (gint x1, gint y1, gint x2, gint y2, guint8 color)
{
	gint lx, ly, dx, dy;
	gint i, j, k;
	lx = abs(x1-x2);
	ly = abs(y1-y2);
	dx = (x1>x2) ? -1 : 1;
	dy = (y1>y2) ? -1 : 1;
	
	if (lx>ly) 
	{
		for (i=x1,j=y1,k=0;i!=x2;i+=dx,k+=ly) 
		{
			if (k>=lx) 
			{
				k -= lx;
				j += dy;
			}
			aff_pixel (i, j, color);
		}
	} 
	else 
	{
		for (i=y1,j=x1,k=0;i!=y2;i+=dy,k+=lx) 
		{
			if (k>=ly) 
			{
				k -= ly;
				j += dx;
			}
		aff_pixel (j, i, color);
		}
	}	
}

void rotation_3d (float *x, float *y, float *z, float alpha, float beta, float gamma)
{
	float nx, ny;
	float cosga = cos (gamma), singa=sin (gamma), cosal=cos (alpha), sinal=sin (alpha);
	nx = cosga * *x - singa * *y;
	ny = cosal * (singa * *x + cosga * *y) - sinal * *z;
	*z = sinal * (singa * *x + cosga * *y) + cosal * *z;
	*x = nx;
	*y = ny;
}

void perspective (float *x, float *y, float *z, int persp, int dist_cam)
{
	float foc_cam = 100;
	*x = *x * (foc_cam + persp) / (dist_cam + persp - *z);
	*y = *y * (foc_cam + persp) / (dist_cam + persp - *z);
}

void grille_3d (float alpha, float beta, float gamma, int persp, int dist_cam, int col_grille_3d, int cx, int cy)
{
	float x, y, z, xres2 = (float) (resx >> 1), yres2 = (float) (resy >> 1);
	gint16 ax = 0, ay = 0, ampli_grille = 0, ix, iy, i, j, nb_x, nb_y;
	nb_x = 32;
	nb_y = 32;
	if (ptr_effect->spectrum == 0 & ptr_effect->lines == 1)
		ampli_grille = 128;
	for (i = 0; i < nb_x; i++)
	{
		for (j = 0; j < nb_y; j++)
		{
			x = RESFACTXF ((i - ((float) nb_x) / 2) * 10);
			y = RESFACTYF ((j - ((float) nb_y) / 2) * 10);
			z = RESFACTXF ((float) (pcm_buffer[i + (nb_x/3)*j])/(1024-ampli_grille));
			rotation_3d (&x, &y, &z, alpha, beta, gamma);
			perspective (&x, &y, &z, persp, dist_cam);
			ix = (int) x;
			iy = (int) y;
			if (j != 0)	{
				droite (ix+cx, iy+cy, ax+cx, ay+cy, col_grille_3d);
			}
			ax = ix;
			ay = iy;
		}
	}
}

void cercle_3d (float alpha, float beta, float gamma, int persp, int dist_cam, int col_grille_3d, int cx, int cy)
{
	float x, y, z, xres2 = (float) (resx >> 1), yres2 = (float) (resy >> 1);
	gint16 ax = 0, ay = 0, ix, iy, i, j, nb_x, nb_y;
	nb_x = 16;
	nb_y = 16;
	for (i = 0; i < nb_x; i++)
	{
		for (j = 0; j < nb_y; j++)
		{
			x = RESFACTXF ((i - ((float) nb_x) / 2) * 30);
			y = RESFACTYF ((j - ((float) nb_y) / 2) * 30);
			z = RESFACTXF ((float) (render_buffer[i + (nb_x/32)*j])/64);
			rotation_3d (&x, &y, &z, alpha, beta, gamma);
			perspective (&x, &y, &z, persp, dist_cam);
			ix = (int) x;
			iy = (int) y;
			if (j != 0)	{
				cercle (ix+cx, iy+cy, 3, col_grille_3d);
				cercle (ax+cx, ay+cy, 3, col_grille_3d);
			}
			ax = ix;
			ay = iy;
		}
	}
}

void do_radial_wave(void)
{
	int i, j, halfheight, halfwidth, shift, col_fleur;
	float y1, old_y1, k, opt, opt_old;
	y1 = pcm_buffer[0]/450;
	col_fleur = 74;
	shift = resy/3.1;
	if (ptr_effect->effect)
		col_fleur = 72;
	if (cosw.i != resx || sinw.i != resx) {
		free(cosw.f);
		free(sinw.f);
		sinw.f = cosw.f = NULL;
		sinw.i = cosw.i = 0;
	}
	if (cosw.i == 0 || cosw.f == NULL) {
	 	float halfPI  = (float)PI/2;
		cosw.i = resx;
		cosw.f = malloc(sizeof(float)*resx);
		for (i=0; i<resx;i++) 
			cosw.f[i] = cos((float)i/resx*PI+halfPI);
	}
	if (sinw.i == 0 || sinw.f == NULL) {
	 	float halfPI = (float)PI/2;
		sinw.i = resx;
		sinw.f = malloc(sizeof(float)*resx);
		for (i=0; i<resx;i++) 
			sinw.f[i] = sin((float)i/resx*PI+halfPI);
	}
	halfheight = resy >> 1;
	halfwidth  = resx >> 1;
	for (i=1 ; i<resx ; i++) {
		old_y1 = y1;
		y1 = pcm_buffer[i]/450;
		if (y1<-50)
			y1 = y1/2;
		for (k=1 ; k<2.6 ; k+=0.4)	{
			opt = shift+(y1/k);
			opt_old = shift+(old_y1/k);
			droite (halfwidth + cosw.f[i-1] * opt_old,
				halfheight + sinw.f[i-1] * opt_old,
				halfwidth + cosw.f[i] * opt,
				halfheight + sinw.f[i] * opt,
				col_fleur);
			droite (halfwidth - cosw.f[i-1] * opt_old,
				halfheight + sinw.f[i-1] * opt_old,
				halfwidth - cosw.f[i] * opt,
				halfheight + sinw.f[i] * opt,
				col_fleur);
		}
	}
	for (k=1 ; k<2.6 ; k+=0.4)	{
			opt = shift+(y1/k);
		droite (halfwidth  + cosw.f[1] * opt,
			halfheight - sinw.f[1] * opt,
			halfwidth  - cosw.f[1] * opt,
			halfheight - sinw.f[1] * opt,
			col_fleur);
		droite (halfwidth  - cosw.f[1] * opt,
			halfheight - sinw.f[1] * opt,
			halfwidth  + cosw.f[1] * opt,
			halfheight - sinw.f[1] * opt,
			col_fleur);
	}
}

void do_tourni_spec(void)
{
	int i, j, k, l, halfheight, halfwidth, col_tourni, amplitude;
	float v, vr, x, y;
	v = 0.1;
	vr = 0.00001;
	col_tourni = 52;
	amplitude = 44;
	j = 20;
	k = ptr_effect->k_put;
	l = 1000;
	halfheight = resy >> 1;
	halfwidth  = resx >> 1;
	if (ptr_effect->effect)
		col_tourni = 99;
	if (ptr_effect->lines == 0 & ptr_effect->spectrum == 1)	{
		j = 4;
		l = 2500;
		amplitude = 88;
	}
	for (i=0 ; i<l ; i++ , k++) {
		x = cos(k/(v*ptr_effect->rot_tourni))*amplitude+(render_buffer[i>>2]/j);
		y = sin(k/(v*0.6))*amplitude+(render_buffer[i>>2]/20);
		aff_pixel (x*cos(k*vr)+y*sin(k*vr)+halfwidth, x*sin(k*vr)-y*cos(k*vr)+halfheight, col_tourni);
	}
	ptr_effect->k_put = k;
}

void do_lines(void)
{
	int lc, lc2lc, reg_lines,larg_lines;
	float pluseffect, coul_lines;
	if (ptr_effect->lines == 1 & ptr_effect->spectrum == 0)	{
		reg_lines = (HEIGHT/2)-3;
		larg_lines = 240;
	}
	else	{
		reg_lines = 77;
		larg_lines = 420;
	}
	for (lc=0 ; lc<WIDTH ; lc++)	{
		for (lc2lc=0 ; lc2lc<7 ; lc2lc++)	{
			coul_lines = 55;
			pluseffect = 0;	
			if (ptr_effect->effect)	{
				coul_lines = 227.4;
				pluseffect = 0.5;
			}
			if (lc2lc==0 || lc2lc==1 || lc2lc==4 || lc2lc==5 || lc2lc==6)
				coul_lines = 400/(3.4+pluseffect);
			aff_pixel(lc, reg_lines+(pcm_buffer[lc]/larg_lines)+lc2lc, coul_lines);
		}
	}
}

void do_spectrum(void)
{
	int lc, lc2lc, lc2lc2lc, l_spect, toujours=0, coul_lines;
	int i=0, too_hard=64, first_decr=FALSE;
	float precision = 71;
	if (ptr_effect->lines == 0 & ptr_effect->spectrum == 1)	{
		val_maxi = 255;
		l_spect = 14;
	}
	else	{
		val_maxi = 127;
		l_spect = 32;
	}
	for (lc=1 ; lc<WIDTH ; lc+=10 , toujours++, too_hard-=i)	{
		for (lc2lc=0; lc2lc<7 ; lc2lc++, precision+=0.08)	{
			for (lc2lc2lc=0 ; lc2lc2lc<7 ; lc2lc2lc++)	{
				coul_lines=7.4+chcol0+precision+(lc2lc*2);
				if (lc2lc2lc==0 || lc2lc2lc==1 || lc2lc2lc==5 || lc2lc2lc==6)
					coul_lines = chcol0+(lc2lc*2);
				aff_pixel(lc+lc2lc, 393-((render_buffer[toujours]/(l_spect+too_hard))&val_maxi)+lc2lc2lc, coul_lines);
			}
		}
		if (!too_hard)
			too_hard = 2;
		if (lc<33)	{
			i = 6;
			first_decr = TRUE;
		}
		else	{
			if (first_decr)	{
				first_decr = FALSE;
				too_hard -= 10;
			}
			i = 2;
		}
	}
}

void do_grille_3d(void)
{
	int dis_col = 322;
	if (ptr_effect->effect)
		dis_col = 72;
	if (ptr_effect->spectrum)
		grille_3d (1.45, 0, compt_grille_3d/30, 200, 110, dis_col, 235, 40);
	else
		grille_3d (1.1, 0, compt_grille_3d/30, 200, 20, dis_col, 235, 80);
	compt_grille_3d += 0.1;
}	

void do_cercle_3d(void)
{
	int dis_col = 382;
	if (ptr_effect->effect)
		dis_col = 99;
	if (ptr_effect->lines)
		cercle_3d (3.6, 0, compt_cercle_3d/30, 200, 80, dis_col, 235, 400);
	else
		cercle_3d (3.2, 0, compt_cercle_3d/30, 200, 80, dis_col, 235, 195);
	compt_cercle_3d -= 0.1;
}	

void what_display(void)
{
	if (ptr_effect->use_3d)	{
		if (ptr_effect->options)	{
			if (ptr_effect->lines)
				do_radial_wave();
			if (ptr_effect->spectrum)
				do_tourni_spec();
		}
		else	{
			if (ptr_effect->lines)
				do_grille_3d();
			if (ptr_effect->spectrum)
				do_cercle_3d();
		}
	}
	else	{
		if (ptr_effect->lines)
			do_lines();
		if (ptr_effect->spectrum)
			do_spectrum();
	}
}

static void func_thread(void)
{
	while (!finished)	{
		SDL_LockSurface(ecran_sdl);
		if (state == old_state)
			state++;
		old_state = state;
		if (changement)	{
			changement = FALSE;
		plazma_change_effect();
		SDL_SetColors(ecran_sdl, colors, 0, TONES);	
		}			
		switch (ptr_effect->effect)	{
		case 0: 
   			do_plasma(ecran_sdl,
			(sin(((double)state)*R[0])+1)/2,
			(sin(((double)state)*R[1])+1)/2,
			(cos(((double)state)*R[2])+1)/2,
			(cos(((double)state)*R[3])+1)/2, tmem);
			break;
		case 1:
   			do_plasma(ecran_sdl,
			(sin(((double)state)*R[0])+1)/2,
			(sin(((double)state)*R[1])+1)/2,
			(cos(((double)state)*R[2])+1)/2,
			(cos(((double)state)*R[3])+1)/2, tmem1);
			break;
		}
		what_display();
		SDL_UnlockSurface(ecran_sdl);
		SDL_Flip(ecran_sdl);
		SDL_PollEvent(&plazma_event);
		if (plazma_event.type == SDL_KEYDOWN)	{
			char *st;
			st=SDL_GetKeyName(plazma_event.key.keysym.sym);
			if (!strcmp(st,"escape"))	{
				SDL_WM_ToggleFullScreen(ecran_sdl);
				mouse = fullscreen?1:0;
				SDL_ShowCursor(mouse);
				fullscreen = fullscreen?FALSE:TRUE;
			}
		}
		if (plazma_event.type == SDL_QUIT)	{
			finished = TRUE;
			in_thread = TRUE;
		}
	}
	if (in_thread)	{
		free(tmem1);
		free(tmem);
		SDL_FreeSurface(ecran_sdl);
		ecran_sdl = NULL;
		printf("Clear memory...\n");
		SDL_Quit();
		printf("Close SDL screen...\n");
	}
}
